
""" A selection of useful algorithms. """

# TODO For backwards compatibility, remove in major version.
from .longest_common_substring import longest_common_substring

__all__ = ['longest_common_substring']
